from tunnel import TunnelMachine


REST_BASE_URL = "https://saucelabs.com/rest/v1/"
DEFAULT_KGP_PORT = 12345


def create_vm(username, password, domains=None, port=DEFAULT_KGP_PORT,
              boost_mode=False, fast_fail_regexps=None, direct_domains=None,
              shared_tunnel=False, squid_opts=None, metadata=None):

    _domains = domains or ["sauce-connect.proxy"]
    print "Domain is %s" % _domains

    # rest_url, user, password, domains, ssh_port, boost_mode, use_ssh,
    # fast_fail_regexps, direct_domains, shared_tunnel, squid_opts, metadata
    tunnel = TunnelMachine(REST_BASE_URL, username, password, _domains, port,
                           boost_mode, fast_fail_regexps, direct_domains,
                           shared_tunnel, squid_opts, metadata)
    tunnel.ready_wait()
    return tunnel



import httplib
import json
import logging
import re
import socket
import time
import urllib2
from base64 import b64encode
from contextlib import closing
from functools import wraps


RETRY_REST_WAIT = 10  # seconds
RETRY_REST_MAX = 3
RETRY_PROVISION_MAX = 10
REST_POLL_WAIT = 1  # seconds

FORMAT = '%(message)s'
logging.basicConfig(format=FORMAT)
logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR) #DEBUG)


class DeleteRequest(urllib2.Request):
    def get_method(self):
        return "DELETE"


class HTTPResponseError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return "HTTP server responded with '%s' (expected 'OK')" % self.msg


class ShutdownRequestedError(Exception):
    pass


class TunnelMachineError(Exception):
    pass


class TunnelMachineProvisionError(TunnelMachineError):
    pass


class TunnelMachineBootError(TunnelMachineError):
    pass


class TunnelMachine(object):

    _host_search = re.compile("//([^/]+)").search

    def __init__(self, rest_url, user, password,
                 domains, ssh_port, boost_mode, use_ssh,
                 fast_fail_regexps, direct_domains, shared_tunnel,
                 squid_opts, metadata=None):
        self.user = user
        self.password = password
        self.domains = set(domains)
        self.ssh_port = ssh_port
        self.boost_mode = boost_mode
        self.use_ssh = use_ssh
        self.fast_fail_regexps = fast_fail_regexps
        self.direct_domains = direct_domains
        self.shared_tunnel = shared_tunnel
        self.squid_opts = squid_opts
        self.metadata = metadata or dict()

        self.reverse_ssh = None
        self.is_shutdown = False
        self.base_url = "%(rest_url)s/%(user)s/tunnels" % locals()
        self.rest_host = self._host_search(rest_url).group(1)
        self.basic_auth_header = {"Authorization": "Basic %s"
                                  % b64encode("%s:%s" % (user, password))}

        self._set_urlopen(user, password)

        for attempt in xrange(1, RETRY_PROVISION_MAX):
            try:
                self._provision_tunnel()
                break
            except TunnelMachineProvisionError, e:
                logger.warning(e)
                if attempt == RETRY_PROVISION_MAX:
                    raise TunnelMachineError(
                        "!! Could not provision tunnel remote VM. Please "
                        "contact help@saucelabs.com.")

    def _set_urlopen(self, user, password):
        # always send Basic Auth header (HTTPBasicAuthHandler was unreliable)
        opener = urllib2.build_opener()
        opener.addheaders = self.basic_auth_header.items()
        self.urlopen = opener.open

    # decorator
    def _retry_rest_api(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            previous_failed = False
            for attempt in xrange(1, RETRY_REST_MAX + 1):
                try:
                    result = f(*args, **kwargs)
                    if previous_failed:
                        logger.info("Connection succeeded")
                    return result
                except (HTTPResponseError,
                        urllib2.URLError, httplib.HTTPException,
                        socket.gaierror, socket.error), e:
                    logger.warning("Problem connecting to Sauce Labs REST API "
                                   "(%s)", str(e))
                    if attempt == RETRY_REST_MAX:
                        raise TunnelMachineError(
                            "Could not reach Sauce Labs REST API after %d "
                            "tries. Is your network down or firewalled?"
                            % attempt)
                    previous_failed = True
                    logger.debug("Retrying in %ds", RETRY_REST_WAIT)
                    time.sleep(RETRY_REST_WAIT)
                except Exception, e:
                    raise TunnelMachineError(
                        "An error occurred while contacting Sauce Labs REST "
                        "API (%s). Please contact help@saucelabs.com." %
                        str(e))
        return wrapper

    @_retry_rest_api
    def _get_doc(self, url_or_req):
        try:
            with closing(self.urlopen(url_or_req)) as resp:
                if resp.msg != "OK":
                    raise HTTPResponseError(resp.msg)
                return json.loads(resp.read())
        except Exception, ex:
            # Running under Jython, JavaException is raised instead of
            # HTTP Exceptions
            raise HTTPResponseError(unicode(ex))

    def kill_tunnel(self, tunnel_id=None):
        if tunnel_id is None:
            tunnel_id = self.id
        for attempt in xrange(1, 4):  # try a few times, then bail
            logger.debug(
                "Shutting down old tunnel remote VM: %s" % tunnel_id)
            url = "%s/%s" % (self.base_url, tunnel_id)
            doc = self._get_doc(DeleteRequest(url=url))
            if (not doc.get('result') or
                not doc.get('id') == tunnel_id):
                logger.warning("Old tunnel remote VM failed to shut "
                               "down.  Status: %s", doc)
                continue
            doc = self._get_doc(url)
            while doc.get('status') not in ["halting", "terminated"]:
                logger.debug(
                    "Waiting for old tunnel remote VM to start "
                    "halting")
                time.sleep(REST_POLL_WAIT)
                doc = self._get_doc(url)
            break

    def _provision_tunnel(self):
        # Shutdown any tunnel using a requested domain
        kill_list = set()
        for doc in self._get_doc(self.base_url + "?full=1"):
            if not doc.get('domain_names'):
                continue
            if set(doc['domain_names']) & self.domains:
                kill_list.add(doc['id'])
        if kill_list:
            logger.info("Shutting down other tunnel remote VMs using "
                        "requested domains")
            for tunnel_id in kill_list:
                self.kill_tunnel(tunnel_id)

        # Request a tunnel machine
        headers = {"Content-Type": "application/json"}
        data = json.dumps(dict(domain_names=list(self.domains),
                               metadata=self.metadata,
                               ssh_port=self.ssh_port,
                               use_caching_proxy=self.boost_mode,
                               use_kgp=not self.use_ssh,
                               fast_fail_regexps=(
                                   self.fast_fail_regexps.split(',') if
                                   self.fast_fail_regexps else None),
                               direct_domains=(
                                   self.direct_domains.split(',') if
                                   self.direct_domains else None),
                               shared_tunnel=self.shared_tunnel,
                               squid_config=(
                                   self.squid_opts.split(',') if
                                   self.squid_opts else None)))
        logger.info("%s" % data)
        req = urllib2.Request(url=self.base_url, headers=headers, data=data)
        doc = self._get_doc(req)
        if doc.get('error'):
            raise TunnelMachineProvisionError(doc['error'])
        for key in ['id']:
            if not doc.get(key):
                raise TunnelMachineProvisionError(
                    "Document for provisioned tunnel remote VM is missing "
                    "the key or value for '%s'" % key)
        self.id = doc['id']
        self.url = "%s/%s" % (self.base_url, self.id)
        logger.info("Tunnel remote VM is provisioned (%s)" % self.id)

    def ready_wait(self):
        """Wait for the machine to reach the 'running' state."""
        previous_status = None
        while True:
            doc = self._get_doc(self.url)
            status = doc.get('status')
            if status == "running":
                break
            if status in ["halting", "terminated"]:
                if doc.get('user_shut_down'):
                    raise ShutdownRequestedError("Tunnel shut down by user "
                                                 "(or another Sauce Connect "
                                                 "process), quitting")
                else:
                    raise TunnelMachineBootError("Tunnel remote VM was shut "
                                                 "down")
            if status != previous_status:
                logger.info("Tunnel remote VM is %s .." % status)
            previous_status = status
            time.sleep(REST_POLL_WAIT)
        self.host = doc['host']
        logger.info("Tunnel remote VM is running at %s" % self.host)

    def shutdown(self):
        if self.is_shutdown:
            return

        if self.reverse_ssh:
            self.reverse_ssh.stop()

        logger.info("Shutting down tunnel remote VM (please wait)")
        logger.debug("tunnel remote VM ID: %s" % self.id)

        try:
            doc = self._get_doc(DeleteRequest(url=self.url))
        except TunnelMachineError, e:
            logger.warning("Unable to shut down tunnel remote VM")
            logger.debug("Shut down failed because: %s", str(e))
            self.is_shutdown = True  # fuhgeddaboudit
            return
        #assert doc.get('ok')
        assert doc.get('id') == self.id

        previous_status = None
        while True:
            doc = self._get_doc(self.url)
            status = doc.get('status')
            if status == "terminated":
                break
            if status != previous_status:
                logger.info("Tunnel remote VM is %s .." % status)
            previous_status = status
            time.sleep(REST_POLL_WAIT)
        logger.info("Finished shutting down tunnel remote VM")
        self.is_shutdown = True

    # Make us usable with contextlib.closing
    close = shutdown

    def check_running(self):
        doc = self._get_doc(self.url)
        if doc.get('status') == "running":
            return
        raise TunnelMachineError(
            "The tunnel remote VM is no longer running. It may have been "
            "shutdown via the website or by another Sauce Connect script "
            "requesting these domains: %s" % list(self.domains))



import ctypes as ct
import ctypes.util


def _do_find_library(name):
    # First just try to find in system library paths
    try:
        lib = ct.CDLL(name, mode=ct.RTLD_GLOBAL)
        if lib:
            return lib
    except:
        pass

    # Next, try ctypes.util.find_library()
    p = ctypes.util.find_library(name)
    if p:
        lib = ct.CDLL(p, mode=ct.RTLD_GLOBAL)
        return lib

    # Next, python library installation path
    import os
    from distutils.sysconfig import get_python_lib
    try:
        lib = ct.CDLL(os.path.join(get_python_lib(), name),
                      mode=ct.RTLD_GLOBAL)
        return lib
    except:
        pass

    # Last, python search path for modules
    import sys
    for p in sys.path:
        try:
            lib = ct.CDLL(os.path.join(p, name), mode=ct.RTLD_GLOBAL)
            return lib
        except:
            pass

    # Out of ideas
    return None


def _find_library(*names):
    lib = None
    for name in names:
        lib = _do_find_library(name)
        if lib is not None:
            return lib
        if not name.startswith("lib"):
            lib = _do_find_library("lib" + name)
            if lib is not None:
                return lib
        for suffix in [".so", ".dylib", ".dll"]:
            if not name.endswith(suffix):
                lib = _do_find_library(name + suffix)
                if lib is not None:
                    return lib
            if not name.startswith("lib") and not name.endswith(suffix):
                lib = _do_find_library("lib" + name + suffix)
                if lib is not None:
                    return lib
    return lib


_lib = _find_library("sauceconnect")
if _lib is None:
    raise Exception("can't find libsauceconnect")

_sc_new = _lib.sc_new
_sc_new.restype = ct.c_void_p
_sc_new.argtypes = None

_sc_free = _lib.sc_free
_sc_free.restype = None
_sc_free.argtypes = [ct.c_void_p]

_SC_PARAM_IS_SERVER = 0x01
_SC_PARAM_KGP_HOST = 0x02
_SC_PARAM_KGP_PORT = 0x03
_SC_PARAM_EXT_HOST = 0x04
_SC_PARAM_EXT_PORT = 0x05
_SC_PARAM_LOGFILE = 0x06
_SC_PARAM_LOGLEVEL = 0x07
_SC_PARAM_MAX_LOGSIZE = 0x08
_SC_PARAM_CERTFILE = 0x09
_SC_PARAM_KEYFILE = 0x0a
_SC_PARAM_LOCAL_PORT = 0x0b
_SC_PARAM_USER = 0x0c
_SC_PARAM_API_KEY = 0x0d

_sc_get = _lib.sc_get
_sc_get.restype = ct.c_int
_sc_get.argtypes = [ct.c_void_p, ct.c_int, ct.c_void_p, ct.c_size_t]

_sc_set = _lib.sc_set
_sc_set.restype = ct.c_int
_sc_set.argtypes = [ct.c_void_p, ct.c_int, ct.c_void_p]

_sc_init = _lib.sc_init
_sc_init.restype = ct.c_int
_sc_init.argtypes = [ct.c_void_p]

_sc_run = _lib.sc_run
_sc_run.restype = ct.c_int
_sc_run.argtypes = [ct.c_void_p]

_sc_stop = _lib.sc_stop
_sc_stop.restype = ct.c_int
_sc_stop.argtypes = [ct.c_void_p]

_SC_STATUS_RUNNING = 0x01
_SC_STATUS_EXITING = 0x02

_sc_status = _lib.sc_status
_sc_status.restype = ct.c_int
_sc_status.argtypes = [ct.c_void_p]


class SauceConnect(object):
    def __init__(self, vm_tunnel, proxy, user, api_key,
                 local_port=8089):
        self.vm_tunnel = vm_tunnel
        self.proxy = proxy
        self._ctx = _sc_new()
        if self._ctx is None:
            raise Exception("failed to create SauceConnect context")
        for p, v in [(_SC_PARAM_IS_SERVER, 0),
                     (_SC_PARAM_KGP_HOST, vm_tunnel.host),
                     (_SC_PARAM_KGP_PORT, vm_tunnel.ssh_port),
                     (_SC_PARAM_EXT_HOST, proxy.host),
                     (_SC_PARAM_EXT_PORT, proxy.port),
                     (_SC_PARAM_LOCAL_PORT, local_port),
                     (_SC_PARAM_USER, user),
                     (_SC_PARAM_API_KEY, api_key),
                     (_SC_PARAM_LOGLEVEL, 2)]:
            if isinstance(v, int):
                i = ct.c_int(v)
                ret = _sc_set(self._ctx, p, ct.cast(ct.pointer(i),
                                                    ct.c_void_p))
            else:
                s = ct.c_char_p()
                s.value = v
                ret = _sc_set(self._ctx, p, ct.cast(s, ct.c_void_p))
            if ret:
                raise Exception("failed to set %d -> %s" % (p, str(v)))
        if _sc_init(self._ctx):
            raise Exception("failed to initialize the SauceConnect engine")

    def __del__(self):
        if self._ctx:
            _sc_free(self._ctx)
            self._ctx = None

    def run(self):
        return _sc_run(self._ctx)

    def stop(self):
        _sc_stop(self._ctx)
        self.proxy.stop()
        self.vm_tunnel.kill_tunnel()



from subprocess import Popen


class SCProxy(object):
    def __init__(self, host="127.0.0.1", port=8123, cmd="scproxy"):
        self.host = host
        self.port = port
        self.cmd = cmd

    def start(self):
        self.proxy = Popen([self.cmd])  # XXX: implement port

    def stop(self):
        rv = self.proxy.poll()
        if rv is None:
            self.proxy.kill()
        return self.proxy.returncode



import threading
import Queue

from os import environ

from proxy import SCProxy
from util import create_vm
from libsauceconnect import SauceConnect

class SC(object):

    def __init__(self, user=environ['SAUCE_USERNAME'], api_key=environ['SAUCE_ACCESS_KEY']):
        self.proxy = SCProxy()
        self.proxy.start()

        vm = create_vm(user, api_key)

        self.sc = SauceConnect(vm, self.proxy, user, api_key)

        self.q = Queue.Queue()


    def _scrunner(self):
        rv = self.sc.run()
        self.q.put(rv)

    def start(self):
        self.t = threading.Thread(target=self._scrunner)
        self.t.daemon = True
        self.t.start()

    def stop(self):
        self.sc.stop()

    def __del__(self):
        if hasattr(self, 'sc'):
            self.sc.stop()

        if hasattr(self, 'proxy'):
            self.proxy.stop()


class SCGuard(SC):
    def __enter__(self):
        self.start()

    def __exit__(self, type, value, traceback):
        self.stop()



#!/usr/bin/env python

from distutils.core import setup

setup(
    name="sauceconnect",
    version="0.0.1",
    description="Python bindings for libsauceconnect",
    author="Sauce Labs",
    author_email="info@saucelabs.com",
    url="http://saucelabs.com/",
    packages=["sauceconnect"],
)
